@@ -27,6 +27,7 @@
 #include <limits>
 
 #include "array.h"
+#include "Dac.h"
 #include "Filter6581.h"
 #include "Filter8580.h"
 #include "Potentiometer.h"
@@ -37,6 +38,78 @@
 namespace reSIDfp
 {
 
+const unsigned int ENV_DAC_BITS = 8;
+const unsigned int OSC_DAC_BITS = 12;
+
+/**
+ * The waveform D/A converter introduces a DC offset in the signal
+ * to the envelope multiplying D/A converter. The "zero" level of
+ * the waveform D/A converter can be found as follows:
+ *
+ * Measure the "zero" voltage of voice 3 on the SID audio output
+ * pin, routing only voice 3 to the mixer ($d417 = $0b, $d418 =
+ * $0f, all other registers zeroed).
+ *
+ * Then set the sustain level for voice 3 to maximum and search for
+ * the waveform output value yielding the same voltage as found
+ * above. This is done by trying out different waveform output
+ * values until the correct value is found, e.g. with the following
+ * program:
+ *
+ *        lda #$08
+ *        sta $d412
+ *        lda #$0b
+ *        sta $d417
+ *        lda #$0f
+ *        sta $d418
+ *        lda #$f0
+ *        sta $d414
+ *        lda #$21
+ *        sta $d412
+ *        lda #$01
+ *        sta $d40e
+ *
+ *        ldx #$00
+ *        lda #$38        ; Tweak this to find the "zero" level
+ *l       cmp $d41b
+ *        bne l
+ *        stx $d40e        ; Stop frequency counter - freeze waveform output
+ *        brk
+ *
+ * The waveform output range is 0x000 to 0xfff, so the "zero"
+ * level should ideally have been 0x800. In the measured chip, the
+ * waveform output "zero" level was found to be 0x380 (i.e. $d41b
+ * = 0x38) at an audio output voltage of 5.94V.
+ *
+ * With knowledge of the mixer op-amp characteristics, further estimates
+ * of waveform voltages can be obtained by sampling the EXT IN pin.
+ * From EXT IN samples, the corresponding waveform output can be found by
+ * using the model for the mixer.
+ *
+ * Such measurements have been done on a chip marked MOS 6581R4AR
+ * 0687 14, and the following results have been obtained:
+ * * The full range of one voice is approximately 1.5V.
+ * * The "zero" level rides at approximately 5.0V.
+ *
+ *
+ * zero-x did the measuring on the 8580 (https://sourceforge.net/p/vice-emu/bugs/1036/#c5b3):
+ * When it sits on basic from powerup it's at 4.72
+ * Run 1.prg and check the output pin level.
+ * Then run 2.prg andadjust it until the output level is the same...
+ * 0x94-0xA8 gives me the same 4.72 1.prg shows.
+ * On another 8580 it's 0x90-0x9C
+ * Third chip 0x94-0xA8
+ * Fourth chip 0x90-0xA4
+ * On the 8580 that plays digis the output is 4.66 and 0x93 is the only value to reach that.
+ * To me that seems as regular 8580s have somewhat wide 0-level range,
+ * whereas that digi-compatible 8580 has it very narrow.
+ * On my 6581R4AR has 0x3A as the only value giving the same output level as 1.prg
+ */
+//@{
+unsigned int constexpr OFFSET_6581 = 0x380;
+unsigned int constexpr OFFSET_8580 = 0x9c0;
+//@}
+
 /**
  * Bus value stays alive for some time after each operation.
  * Values differs between chip models, the timings used here
@@ -54,7 +127,7 @@ namespace reSIDfp
  * [1]: http://sourceforge.net/p/vice-emu/patches/99/
  * [2]: http://noname.c64.org/csdb/forums/?roomid=11&topicid=29025&showallposts=1
  */
- //@{
+//@{
 int constexpr BUS_TTL_6581 = 0x01d00;
 int constexpr BUS_TTL_8580 = 0xa2000;
 //@}
@@ -152,14 +225,42 @@ void SID::setChipModel(ChipModel model)
 
     this->model = model;
 
-    // calculate waveform-related tables, feed them to the generator
+    // calculate waveform-related tables
     matrix_t* tables = WaveformCalculator::getInstance()->buildTable(model);
 
-    // update voice offsets
+    // calculate envelope DAC table
+    {
+        Dac dacBuilder(ENV_DAC_BITS);
+        dacBuilder.kinkedDac(model);
+
+        for (unsigned int i = 0; i < (1 << ENV_DAC_BITS); i++)
+        {
+            envDAC[i] = static_cast<float>(dacBuilder.getOutput(i));
+        }
+    }
+
+    // calculate oscillator DAC table
+    const bool is6581 = model == MOS6581;
+
+    {
+        Dac dacBuilder(OSC_DAC_BITS);
+        dacBuilder.kinkedDac(model);
+
+        const double offset = dacBuilder.getOutput(is6581 ? OFFSET_6581 : OFFSET_8580);
+
+        for (unsigned int i = 0; i < (1 << OSC_DAC_BITS); i++)
+        {
+            const double dacValue = dacBuilder.getOutput(i);
+            oscDAC[i] = static_cast<float>(dacValue - offset);
+        }
+    }
+
+    // set voice tables
     for (int i = 0; i < 3; i++)
     {
-        voice[i]->envelope()->setChipModel(model);
-        voice[i]->wave()->setChipModel(model);
+        voice[i]->setEnvDAC(envDAC);
+        voice[i]->setWavDAC(oscDAC);
+        voice[i]->wave()->setModel(is6581);
         voice[i]->wave()->setWaveformModels(tables);
     }
 }
